/* RT user.

        Copyright (c) 2002 Roger Larsson <roger.larsson@norran.net>

    This program is free software; you can redistribute it and/or
    modify it under the terms of version 2 of the GNU General Public
    License as published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Thanks to autor of KSysGuard Chris Schlaeger for borrowed code...
*/

#include <sys/types.h>
#include <sched.h>
#include <sys/mman.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#define __USE_GNU
#include <dlfcn.h>

#include "request.h"

/*
 * original C library functions
 */

#define DEF_PRELOAD(type, name, params) \
	typedef type (*orig_##name##_ptr)params; \
	static orig_##name##_ptr orig_##name; \
	type monitor_##name params;
#define INIT_PRELOAD(name) \
	orig_##name = (orig_##name##_ptr)dlsym(RTLD_NEXT,#name);

DEF_PRELOAD(int, mlockall, (int flags))
DEF_PRELOAD(int, mlock, (const void *addr, size_t len))
DEF_PRELOAD(int, sched_get_priority_max,(int policy))
DEF_PRELOAD(int, sched_get_priority_min,(int policy))
DEF_PRELOAD(int, sched_setscheduler,(pid_t pid, int policy, const struct sched_param *p))


static int monitor_sched_setscheduler(pid_t pid, int policy, const struct
       sched_param *p);
static int isRT(pid_t pid);

static int rt_debug = 0;
static int rt_ignore = 0;
static int rt_min_prio = 0;
static int rt_max_prio = 0;
static int rt_init = 0;
#define CHECK_INIT() if(!rt_init) rt_doinit();

/*
 * Initialization - maybe this should be either be a startup only called
 * routine, or use pthread locks to prevent strange effects in multithreaded
 * use (however it seems highly unlikely that an application would create
 * multiple threads before even using one of redirected the system functions
 * once).
 */
static void rt_doinit()
{
	const char *env;
	rt_init = 1;

	/* debugging? */
	env = getenv("RTS_VERBOSE");
	rt_debug = env && !strcmp(env,"1");

	env = getenv("RTS_IGNORE");
	rt_ignore = env ? 1 : 0;

	if (sizeof(struct request) > REQUEST_SIZE) {
		fputs("Fatal: Request bigger than reserved space", stderr);
		exit(1);
	}

	/* resolve original symbols */
	/* orig_F = (F_ptr)dlsym(RTLD_NEXT,"F"); */
	INIT_PRELOAD(mlockall)
	INIT_PRELOAD(mlock)
	INIT_PRELOAD(sched_get_priority_max)
	INIT_PRELOAD(sched_get_priority_min)
	INIT_PRELOAD(sched_setscheduler)

	rt_min_prio = orig_sched_get_priority_min(SCHED_RR) + 1;
	rt_max_prio = orig_sched_get_priority_max(SCHED_RR)/2 - 1;
}

int mlock(const void *addr, size_t len)
{
	CHECK_INIT();
	return 0; /* Assume sucess! */
}

int mlockall(int flags)
{
	CHECK_INIT();
	return 0; /* Assume sucess! */
}

int sched_setscheduler(pid_t pid, int policy, const struct
       sched_param *p)
{
	CHECK_INIT();

	if (pid == 0)
	{
		pid = getpid();
	}

	switch (policy)
	{
		case SCHED_OTHER:
			break;
		case SCHED_FIFO:
			policy = SCHED_RR; /* prefere Round Robin to avoid starvation */
			/* no break */
		case SCHED_RR:
			break;
		default:
			errno = EINVAL;
			return -1;
	}

	if (!p) {
		errno = EINVAL;
		return -1;
	}

	if (p->sched_priority < rt_min_prio || p->sched_priority > rt_max_prio) {
		/* out of limited range */
		errno = EINVAL;
		return -1;
	}

	return monitor_sched_setscheduler(pid, policy, p);
}

int sched_get_priority_min(int policy)
{
	CHECK_INIT();

	switch (policy)
	{
		case SCHED_OTHER:
			return orig_sched_get_priority_min(policy);
		case SCHED_FIFO:
			policy = SCHED_RR; /* prefere Round Robin to avoid starvation */
			/* no break */
		case SCHED_RR:
			if (rt_debug)
			    printf("RTS rt_min_prio = %d\n", rt_min_prio);
			return rt_min_prio;
		default:
			errno = EINVAL;
			return -1;
	}
}

int sched_get_priority_max(int policy)
{
	CHECK_INIT();

	switch (policy)
	{
		case SCHED_OTHER:
			return orig_sched_get_priority_max(policy);
		case SCHED_FIFO:
			policy = SCHED_RR; /* prefere Round Robin to avoid starvation */
			/* no break */
		case SCHED_RR:
		        if (rt_debug)
			    printf("RTS rt_max_prio = %d\n", rt_max_prio);
			return rt_max_prio;
		default:
			errno = EINVAL;
			return -1;
	}
}

static int monitor_sched_setscheduler(pid_t pid, int policy, const struct
       sched_param *p)
{
	struct request request = {sizeof(struct request), 0, pid, policy, p->sched_priority};
	FILE *reqf;

	if (rt_ignore)
		return 0;

	printf("rts: Warning: As long as no rt_monitor runs, execution will sleep here...\n");
	reqf = fopen(RTREQUEST_FILE, "w");
	if (reqf == NULL) {
	    perror("fopen");
	    return errno;
	}

	if (rt_debug)
	{
	    printf("from policy %d\n", sched_getscheduler(pid));
	    printf("Wait...\n"); /* Has to be here to get some chance to display... */
	}

	fwrite(&request, 32, 1, reqf);
	fclose(reqf); /* important! (maybe flush?) */

	/*
	 *  Feedback?
	 */

	return 0; /* Assume sucess! */
}

static int isRT(pid_t pid)
 {
     int sched_class = sched_getscheduler( pid);
     if (sched_class == -1) {
	 fprintf(stderr, "Pid %d Exited?\n", pid);
	 return 0;
     }

     return sched_class != SCHED_OTHER;
 }

